# Instant Animator

Milease provides a series of extension functions for the `object` type, including `Milease`, `MileaseTo`, `MileaseAdditive`, and more.

## Creating an Instant Animator

### Method 1: Using DSL for Generation (Recommended)

Compared to traditional generation methods, we’ve observed that developers often need to specify an easing function. However, the easing function is located further down in the optional parameters list, making it less convenient for developers to clearly observe the start value, end value, animation duration, and start time—especially in editors without `Inlay Hints` support.

We use `T4 templates` to generate a series of animation extension functions for `object`, which follow the pattern:

```c#
MilInstantAnimator object.M[EasingFunctionName][EasingBranchName]<T, E>(Expression<Func<T, E>> mbExpr, AniExpression<E> aniExpr);
```

or

```c#
MilInstantAnimator object.M[EasingFunctionName][EasingBranchName]<T, E>(Expression<Func<T, E>> mbExpr, E from, E to);
```

For example, when you need to use `Quad` `Out` easing, the corresponding function is:

```c#
MilInstantAnimator object.MQuadOut<T, E>(Expression<Func<T, E>> mbExpr, E from, E to);
```

Additionally, we provide a series of operator overloads, allowing you to generate an animator/animator segment in this manner.

When you need to use it as a complete animator, we recommend the following syntax:

```c#
// Both approaches are valid
AudioSource.MQuadOut(x => x.volume, [StartTime] + [Duration] / [StartValue].To([EndValue])).Play();
([StartTime] + [Duration] / AudioSource.MQuadOut(x => x.volume, [StartValue], [EndValue])).Play();
```

When transitioning the volume from the current volume to a target volume (instead of starting from a specified volume), you can use the following syntax:

```c#
// Both approaches are valid
AudioSource.MQuadOut(x => x.volume, [StartTime] + [Duration] / [TargetValue].ToThis()).Play();
([StartTime] + [Duration] / AudioSource.MQuadOut(x => x.volume, [TargetValue].ToThis())).Play();
```

When you need to use it as part of an animator segment composition, we recommend the following syntax:

```c#
[StartTime] + [Duration] / AudioSource.MQuadOut(x => x.volume, [StartValue], [EndValue])
```

Or the following:

```c#
[StartTime] + [Duration] / AudioSource.MQuadOut(x => x.volume, [TargetValue])
```

### Method 2: Using the Traditional Generation Method

```c#
MilInstantAnimator object.Milease(string memberName, object startValue, object toValue, float duration);
```

This function creates an animator that transitions `object.memberName` from `startValue` to `toValue` over the `duration`.

`duration` is in seconds, as all time parameters in Milease are.

```c#
MilInstantAnimator object.MileaseTo(string memberName, object toValue, float duration);
```

This function creates an animator that transitions `object.memberName` from its current value to `toValue` over the `duration`.

```c#
MilInstantAnimator object.MileaseAdditive(string memberName, object startValue, object toValue, float duration);
```

This function creates an animator that transitions `object.memberName` from `current value + startValue` to `current value + toValue` over the `duration`.

The animator will **not** start playing immediately after creation.

## Limitations of the Instant Animator

* `memberName` is case **sensitive**.

* The specified member must be writable, public, and non-static.

* The specified member can be a field or a property.
* The specified member type must have one of the following characteristics:
  * Primitive type, such as bool, int, float, double, string, etc.
  * Serializable type with overloaded operators: `type + type`, `type - type`, `type * float`

## Lifecycle of the Instant Animator

By default, it will stop when the scene is destroyed. You can change this behavior using the following code to let it continue playing when the scene is unloaded:

```
animator.DontStopOnLoad();
```

However, please note that an exception will occur if the object being manipulated by the animator is destroyed due to the scene destruction.

Therefore, you may need to manually stop the animator when the object is destroyed:

```c#
animator.Stop();
```

The animator does not require manual resource disposal.

## Reset Modes of the Instant Animator

The reset modes of the instant animator include:

1. Reset to Original State **(default)**

   When resetting, the animator will revert any modifications it made to the object, restoring it to the state it was in before the animator's modifications.

2. Reset to the Initial State Set by the Animator

   When resetting, the object's state will be restored to the earliest state defined on the animator's timeline.

Animations created with `MileaseTo` and `MileaseAdditive` cannot use reset mode 2; they will forcibly use reset mode 1, as they do not have an initial state defined in their settings.

To change the animator's reset mode:

```c#
animator.UsingResetMode(mode);
```

When using reset mode 2, the animator will reset the object before playing the animation.

## Using the Instant Animator

To start playing the animator:

```c#
animator.Play();
```

To pause the animation:

```c#
animator.Pause();
```

To stop the animation:

```c#
animator.Stop();
```

To reset the state:

```c#
animator.Reset(mode);
```

## Custom Animation Handler Functions

Implement the following delegate to create custom animations:

```c#
delegate void MileaseHandleFunction(MilHandleFunctionArgs e);
```

Call `e.GetTarget<T>()` to get the object being animated, and `e.Progress` is the eased progress of the animation, ranging approximately from [0.0, 1.0].

To create a custom animation with `Milease`:

```c#
MilInstantAnimator object.Milease(MileaseHandleFunction animationFunction, MileaseHandleFunction resetFunction, float duration);
```

`resetFunction` can be null and will be called when the animator resets.

## Keyframe Events

Construct keyframe events using the `Action` extension function:

```c#
MilInstantAnimator Action.AsMileaseKeyEvent(float delay = 0);
```

## Nested Complex Animations

Define nested animations using functions like `animator.While` and `animator.Then`, for example:

```c#
var spriteRenderer = GetComponent<SpriteRenderer>();

Animation =
    MilInstantAnimator.Start(
        1f / transform.Milease(x => x.position,new Vector3(0f, 0f, 0f), new Vector3(1f, 1f, 0f))
    )
    .And(
        0.5f + 1f / spriteRenderer.Milease(x => x.color, Color.white, Color.red)
    )
    .Then(
        1f / transform.MBounce(x => transform.localScale, new Vector3(1f, 1f, 1f), new Vector3(2f, 2f, 2f))
    ).Delayed(1f)
    .Then(
        1f / Text.MLinear(x => x.text, "Start!", "Finish!")
    )
    .Then(
        Text.Milease((e) =>
        {
            var text = e.GetTarget<TMP_Text>();
            text.text = $"Hide after {((1f - e.Progress) * 2f):F1}s...";
        }, null,2f, 0f, EaseFunction.Linear)
    )
    .Then(
        Text.gameObject.Milease(HandleFunction.Hide, HandleFunction.AutoActiveReset(Text.gameObject), 0f)
    )
    .Then(
        1f / transform.Milease(x => x.position, Vector3.zero.ToThis())
    );
```

To play a specified animation concurrently with the previous one:

```c#
MilInstantAnimator animator.While(MilInstantAnimation animation);
```

To play a set of specified animations concurrently with the previous one:

```c#
MilInstantAnimator animator.While(params MilInstantAnimation[] animations);
```

To play a specified animation after the previous animation/group has finished:

```c#
MilInstantAnimator animator.Then(MilInstantAnimation animation);
```

To play a set of specified animations after the previous animation/group has finished:

```c#
MilInstantAnimator animator.Then(params MilInstantAnimation[] animations);
```

To play a series of animations in sequence after the previous animation/group has finished:

```c#
MilInstantAnimator animator.ThenOneByOne(params MilInstantAnimation[] animations);
```

To set a delay before each animation in the previous animation/group plays:

```c#
MilInstantAnimator animator.Delayed(float time);
```
